package devmgr.api.rpc;

/**************************************************************************
 *                                                                         *
 *  Copyright 2003-2011 NetApp, Inc. All Rights Reserved.                  *
 *                                                                         *
 *  This file is confidential and a trade secret of NetApp, Inc.  The      *
 *  receipt of or possession of this file does not convey any rights to    *
 *  reproduce or disclose its contents or to manufacture, use, or sell     *
 *  anything it may describe, in whole, or in part, without the specific   *
 *  written consent of NetApp, Inc.                                        *
 *                                                                         *
 **************************************************************************/

import devmgr.api.runtime.ApiDefinition;
import devmgr.api.runtime.Decoder;
import devmgr.api.runtime.XdrMessage;
import devmgr.api.symbol.SYMbolAPI;
import devmgr.api.symbol.SYMbolAPIClientV1;
import devmgr.api.symbolutils.SYMbolAuthGenerator;

import java.io.IOException;
import java.net.*;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.Date;
import java.util.UUID;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicLong;
/**
 * A Non Blocking (or blocking), but NIO based client.  This class is not thread safe.  One call at a time!!!!!
 */
public class NIORPCClient implements AutoCloseable, AbstractRPCControllerClient {

    private static AtomicLong symbolCallCounter = new AtomicLong();

    public static long getSymbolCallCount() {
        return symbolCallCount;
    }

    private static long symbolCallCount;

    public static final SYMbolAPI api = new SYMbolAPI();

	SelectManager mgr;
    final XdrProtocolHelper protocolHelper;

    private int nextXid;

    private static final int DEFAULT_TIMEOUT = -1;

    final private String uuid; //random unique ID for this client
    private SocketChannel m_Channel = null;

    AtomicLong lastTouched=new AtomicLong();

    final Object callSync=new Object();

    //current call state
    int xid;

    byte[] header = new byte[4];
    int bodyLen = -1;
    int lengthCode=0;
    int amountOfHeaderRead = 0;
    int amountOfBodyRead = 0;

    XdrMessage bodyBuffer;
    byte[] currentBodySegment;
    int timeout=10; //time in seconds
    int userDefinedTimeout=DEFAULT_TIMEOUT;

    //called when the connection succeeds or fails, this being null means no current command in progress
    protected NIOClientCallback currentCmdCallback;
    private int proc; //current call proc

    private Throwable lasrError;
    private String controllerRef;
    private Throwable creationStack;
    public NIORPCClient(SelectManager mgr) throws RPCError, IOException {
        this(mgr,null);

    }
    /*
     * If channel is null, one is opened, otherwise the passed in channel is used (Assuming its already open)
     * @exception RPCError
     *                If any RPC error occurs.
     * @exception IOException
     *                If any socket error occurs.
     */
	public NIORPCClient(SelectManager mgr,SocketChannel channel) throws RPCError, IOException {
        creationStack=new Throwable("who created me");
        this.uuid=UUID.randomUUID().toString();
        protocolHelper=new XdrProtocolHelper(api);
        if (channel == null){
            m_Channel = SocketChannel.open();
            m_Channel.setOption(StandardSocketOptions.TCP_NODELAY,Boolean.TRUE);
            m_Channel.configureBlocking(false);
        }else{
            m_Channel=channel;
        }
        this.mgr=mgr;
        mgr.registerClient(this);


        // Set up the transaction ID sequencer using the current time
        // of day value from the system clock.
        nextXid = (int) ( (new Date()).getTime() / 1000 );
	}
    /**
     * Obtain the next valid transaction ID value for use with this RPC
     * client object.
     * @return The next transaction ID value to be used by this client.
     */
    private synchronized int getNextXid()
    {
        return nextXid++;
    }


    @Override
    public void setTypeOverrides(Decoder.TypeOverrideTable typeOverrides) {
        this.protocolHelper.setTypeOverrides(typeOverrides);
    }

    /**
     * Synchnously Connects the client, when the call ends
     * @param server symbol server address
     * @throws Throwable
     */
    public void connect(InetAddress server) throws Exception {
        connect(server, SYMbolAPI.PORT);
    }

    /**
     * Synchnously Connects the client, when the call ends
     * @param server symbol server address
     * @param port SYMbol port
     * @throws Throwable
     */
    public void connect(InetAddress server, int port) throws Exception{

        final BlockingCallbackHandler callback=new BlockingCallbackHandler();
        connect(server, port, callback);
        synchronized (callback){
            while (!callback.isDone()){
                callback.wait(500);
            }
        }
        if (callback.lastError != null){
            throw callback.lastError;
        }
    }
    public void connect(InetAddress server, NIOClientCallback callback) {
        connect(server,SYMbolAPI.PORT,callback);
    }
    /**
     * Asyncrounously Connects the client.  Methods on callback indicate when the connection is done
     * @param server
     * @param callback
     */
	public void connect(InetAddress server,int port, NIOClientCallback callback) {
        lastTouched.set(System.currentTimeMillis());
        InetSocketAddress addr = new InetSocketAddress(server, port);
        this.currentCmdCallback=callback;
        try{
            proc=-1;
            m_Channel.connect(addr);
        }catch (Exception t){

            lasrError=t;
            if (currentCmdCallback != null){
                currentCmdCallback.onError(t);
            }
        }
	}


    @Override
    public int hashCode() {
        return uuid.hashCode();
    }



    @Override
    public boolean equals(Object obj) {
        boolean ret=false;
        if (obj instanceof NIORPCClient){
            NIORPCClient x=(NIORPCClient)obj;
            ret=uuid.equals(x.uuid);
        }
        return ret;
    }



    public SocketChannel getChannel() {
        return m_Channel;
    }



    public long getLastTouched() {
        return lastTouched.get();
    }

    /**
     * DESCRIPTION: <BR>
     * Finalizes objects just prior to garbage collection. <BR>
     * <BR>
     * NOTES/ASSUMPTIONS: <BR>
     * None. <BR>
     *
     * @exception Throwable
     *                If any underlying exception occurs.
     */
    @Override
    protected void finalize() throws Throwable {
        try{
            close();
        }
        finally {
            super.finalize();
        }
    }

    public void setTimeout(int seconds)  {
        userDefinedTimeout=seconds;
    }
    /**
     * Rreturns timeout in seconds
     * @return
     */
    public int getTimeout(){
        return timeout;
    }

    /**
     * DESCRIPTION: <BR>
     * Closes the underlying network connection. <BR>
     * <BR>
     * NOTES/ASSUMPTIONS: <BR>
     * None. <BR>
     *
     * @exception IOException
     *                If an I/O error occurs.
     */
    @Override
    public void close() throws IOException {
        mgr.deregisterClient(this);
        if (m_Channel != null){
            m_Channel.close();
            m_Channel = null;
        }

        return;
    }

    /**
     * DESCRIPTION: <BR>
     * Returns the IP address of the client end of the RPC connection. <BR>
     * <BR>
     * NOTES/ASSUMPTIONS: <BR>
     * None. <BR>
     *
     * @return Client-side IP address.
     */
    public InetAddress getLocalAddress() {
        return m_Channel.socket().getLocalAddress();
    }



    /**
     * DESCRIPTION: <BR>
     * This method returns the server's address that this object is talking to. <BR>
     * <BR>
     * NOTES/ASSUMPTIONS: <BR>
     * None.
     * @return InetAddress
     */
    public InetAddress getDestinationIP() {
        return m_Channel.socket().getInetAddress();
    }

    /**
     * Synnc calls the proc by name
     * @param proc proc name
     * @param args
     * @return
     * @throws Throwable
     */
    @Override
    public <T> T call(String proc, Object args) throws Exception{
        ApiDefinition.ProcTableEntry x = this.api.getProcByName(proc);
        return call(x.procNumber, args);
    }
    /**
     * Synchrousnly makes a XDR call to the array or proxy
     * @param proc
     * @param args
     * @return  The results of the call
     * @throws Throwable
     */
    @Override
    public <T> T call(int proc, Object args) throws Exception{
        final BlockingCallbackHandler callback=new BlockingCallbackHandler<T>();
        call(proc, args,callback);
        synchronized (callback){
            while (!callback.isDone()){
                callback.wait(500);
            }
        }
        if (callback.lastError != null){
            throw callback.lastError;
        }
        return (T)callback.lastResults;
    }

    /**
     * ASync calls the proc by name
     * @param proc
     * @param args
     * @param callback
     * @return
     */
    @Override
    public Object call(String proc, Object args, NIOClientCallback callback){
        ApiDefinition.ProcTableEntry x = this.api.getProcByName(proc);
        return call(x.procNumber,args,callback);
    }

    /**
     * Asynchnously makes a XDR call to the array or proxy
     *
     * @return the ID for the call
     */
    @Override
    public int call(int proc, final Object args, NIOClientCallback callback){
        assert (lasrError==null);
        if (callback == null){
            throw new IllegalArgumentException("Callback cannot be null");
        }
        if (isActiveCall()){
            throw new IllegalStateException("Cannot run concurrent calls on a single client");
        }
        lastTouched.set(System.currentTimeMillis());
        symbolCallCount = symbolCallCounter.incrementAndGet();
        this.proc=proc;
        ApiDefinition.ProcTableEntry entry = api.getProcByNumber(proc);
        timeout = userDefinedTimeout == DEFAULT_TIMEOUT ? (entry.suggestedTimeout*2) : userDefinedTimeout;

        xid = getNextXid();
        //System.out.println("***NEW XID " + proc + " / " +xid);
        this.currentCmdCallback=callback;
        try{
            XdrMessage callMsg = protocolHelper.buildRequestMsg(proc, xid, args);
            resetCallBuffers();
            sendMsg(callMsg);
        }catch (Exception e){
            handleError(e);
        }

        return xid;
    }
    public boolean isConnected(){
        return m_Channel.isConnected();
    }


    private void resetCallBuffers() {
        // reset call buffers
        amountOfHeaderRead = 0;
        amountOfBodyRead = 0;
        bodyLen = -1;
        lengthCode=0;
        currentBodySegment=null;
        bodyBuffer = new XdrMessage();

    }

    /**
     * Called by the select channel manager to handle bytes available from this
     * channel
     *
     * @throws IOException
     */
    private void processBytes() throws IOException {
        long debugTotal=0;
        ByteBuffer readBuf = ByteBuffer.allocate(3096);
        int amtRead=m_Channel.read(readBuf);
        if (amtRead <0){
            handleError(new Exception("Unexpected Channel Close"));
        }
        byte[] buf=readBuf.array();
        while (amtRead > 0) {
            debugTotal+=amtRead;
            int amtToProcess = amtRead;
            // loop on the amount to read to process header and body if we dont
            // have all the heaer yet.
            while ( (amtToProcess > 0)&&  isActiveCall()) {
                if (amountOfHeaderRead != header.length) {
                    int n = Math.min(amtToProcess,
                            (header.length - amountOfHeaderRead));
                    int srcPos = amtRead - amtToProcess;
                    System.arraycopy(buf, srcPos, header, amountOfHeaderRead, n);
                    amountOfHeaderRead += n;
                    amtToProcess -= n;
                    if (amountOfHeaderRead == header.length){
                        // if we have read the header but not yet allocated the
                        // body, do it now
                        lengthCode = (((int) (header[0]) & 0xFF) << 24)
                                | (((int) (header[1]) & 0xFF) << 16)
                                | (((int) (header[2]) & 0xFF) << 8)
                                | (((int) (header[3]) & 0xFF) << 0);
                        bodyLen=lengthCode & 0x7FFFFFFF;
                        currentBodySegment=new byte[bodyLen];
                    }
                }
                // if did not eat all the bits in the header, keep going on the
                // body
                if (amtToProcess > 0) {// NOW READ THE BODY
                    // now process the body of the rest of the remaining bits as
                    // message body
                    int srcPos = amtRead - amtToProcess;
                    int lenToUse=Math.min(bodyLen-amountOfBodyRead, amtToProcess);
                    System.arraycopy(buf, srcPos, currentBodySegment,amountOfBodyRead, lenToUse);
                    amountOfBodyRead += lenToUse;
                    amtToProcess -=lenToUse;
                    if (amountOfBodyRead == bodyLen){
                        //reset the headers to read next chunk
                        amountOfHeaderRead=0;
                        bodyLen=-1;
                        amountOfBodyRead=0;
                        bodyBuffer.addBuffer(currentBodySegment);
                    }
                }

            }
            readBuf = ByteBuffer.allocate(3096);
            amtRead=m_Channel.read(readBuf);
            buf=readBuf.array();

        }
    }

    public void setAuthGenerator(RPCAuthGenerator generator) {
        protocolHelper.setAuthGenerator(generator);
    }


    public Throwable getLasrError() {
        return lasrError;
    }
    // ******************************* T A K E N F R O M T C P C H A N N E L C L A S S *************************************************************************
    /**
     * DESCRIPTION: <BR>
     * Sends an outbound RPC message, which is provided in the form of an
     * XDRMessageIOV by the caller. <BR>
     * <BR>
     * NOTES/ASSUMPTIONS: <BR>
     * None. <BR>
     *
     * @param msg
     *            The message to be sent.
     * @exception RPCError
     *                If an RPC error occurs.
     * @exception IOException
     *                If any I/O error occurs.
     */
    public void sendMsg(XdrMessage msg) throws RPCError, IOException {
        int totalSize = msg.getMessageSize();
        int amtDone = 0;
        int lengthCode;
        int nextBuf = 0;
        int amt;
        byte[] buf;
        byte[] header = new byte[4];

        while (amtDone < totalSize) {
            buf = msg.getBuffer(nextBuf++);
            amt = Math.min(buf.length, totalSize - amtDone);

            lengthCode = amt;
            if ((amtDone + amt) >= totalSize) {
                lengthCode |= 0x80000000;
            }

            // set up the TCP record header
            header[0] = (byte) (lengthCode >>> 24);
            header[1] = (byte) (lengthCode >>> 16);
            header[2] = (byte) (lengthCode >>> 8);
            header[3] = (byte) (lengthCode >>> 0);

            // write the header and data to the output stream
            // m_Out.write(header);
            // m_Out.write(buf, 0, amt);
            ByteBuffer writeBuf = ByteBuffer.allocate(4);
            writeBuf.put(header);
            writeFully(writeBuf);
            writeBuf = ByteBuffer.allocate(amt);
            writeBuf.put(buf, 0, amt);
            writeFully(writeBuf);
            amtDone += amt;
        }

        // force transmission of any buffered data
        m_Channel.socket().getOutputStream().flush();

        return;
    }
    private void writeFully(ByteBuffer writeBuf) throws IOException {
        int n;
        writeBuf.rewind();
        n= m_Channel.write(writeBuf);
        int remaining = writeBuf.remaining();
        while (remaining>0){
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
            n= m_Channel.write(writeBuf);
            remaining=writeBuf.remaining();

        }
    }

    public int getXid() {
        return xid;
    }

    private void endCurrentCommand(){
        currentCmdCallback=null;
    }
    //*********** P A C K A G E    L E V E L   C A L L B A C K S from SELECTOR ************

    synchronized void handleTimeout() {
        lasrError=new Exception("Timeout");
        if (currentCmdCallback != null){
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            NIOClientCallback tmp=currentCmdCallback;
            endCurrentCommand();
            tmp.onTimeout();
        }else{
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            endCurrentCommand();
        }


    }

    synchronized void handleError(Exception t) {
        lasrError=t;
        if (currentCmdCallback != null){
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            NIOClientCallback tmp=currentCmdCallback;
            endCurrentCommand();
            tmp.onError(t);
        }else{
            //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
            endCurrentCommand();
        }

    }
    synchronized void handleConnectSocketSignal() throws IOException{
        if (m_Channel.isConnectionPending()) {
            try{
                m_Channel.finishConnect();
                //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
                NIOClientCallback tmp=currentCmdCallback;
                endCurrentCommand();
                tmp.onSuccess(null);
            }catch (Exception t){
                lasrError=t;
                NIOClientCallback tmp=currentCmdCallback;
                endCurrentCommand();
                //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
                tmp.onError(t);
            }

        }else{
            //not sure what this means to get here, will let it go for now
        }
    }

    synchronized public void handleReadSignal() throws IOException {
        processBytes();
        synchronized (callSync) {
            if (currentCmdCallback == null){
                return;
            }
        }
        boolean doneCode=((lengthCode & 0x80000000) != 0) ;
        boolean allDataFetched=bodyLen == -1 ? true : amountOfBodyRead == bodyLen;
        boolean done=doneCode && allDataFetched;
        if (done ) {
            //System.out.println("DONE ON " + proc + " / " + xid);
            //grab a copy of the command and null out the current command, so its handlers can make new calls
            NIOClientCallback cmdToUse=currentCmdCallback;
            try {
                Object result = protocolHelper.processReplyMsg(proc, bodyBuffer, xid);
                //Dont get fancy and try to refactor and make only 1 call toendCurrentCommand, it is this way to support multi-threaded world
                endCurrentCommand();
                cmdToUse.onSuccess(result);
            } catch (Exception e) {
                lasrError=e;

                //Dont get fancy and try to refactor and make only 1 call to endCurrentCommand, it is this way to support multi-threaded world
                endCurrentCommand();
                cmdToUse.onError(e);
            }

        }
    }

    public boolean isActiveCall() {
        boolean ret;
        synchronized (callSync){
            ret=currentCmdCallback != null;
        }
        return ret;
    }

    public void setControllerRef(String controllerRef) {
        this.controllerRef = controllerRef;
    }

    public String getControllerRef() {
        return controllerRef;
    }


    /************** INNER CLASSES *********************************
     * Blocking Callback Handler
     * This class is used if the user does not specify a callback, making the calls a blocking call
     */
    static class BlockingCallbackHandler<T> implements NIOClientCallback<T>{
        Exception lastError;
        T lastResults;
        volatile boolean done;
        @Override
        public void onTimeout() {
            lastError=new TimeoutException("timeout");
            signal();
        }

        boolean isDone() {
            return done;
        }

        private void signal() {
            done=true;
            synchronized (this){
                notifyAll();
            }
        }

        @Override
        public void onError(Exception t) {
            lastError=t;
            signal();
        }

        @Override
        public void onSuccess(T results) {
            lastResults=results;
            signal();
        }
    }

    public Throwable getCreationStack() {
        return creationStack;
    }
}
