package devmgr.api.rpc;

import devmgr.api.rpc.protocol.RPCOpaqueAuth;
import devmgr.api.runtime.XdrMessage;

/**
 * This interface is used by the RPC layer to specify interactions
 * between the underlying RPC code and the higher level object that
 * supplies the authentication data for RPC messages.
 */
public interface RPCAuthGenerator
{
    /**
     * Return the "initial" authentication information object for an RPC
     * request message.  If the authentication generator creates information
     * that does not depend on the content of the message, then the full
     * information can be returned as the result of this operation; the
     * setFinalAuthData() method can be a null operation such cases.
     * Otherwise, a partially-filled object of the appropriate size can be
     * returned here, and then the setFinalAuthData() method used to overlay
     * the final data after the message has been fully constructed.
     * @return The initial authentication information.
     * @exception RPCError If any RPC error occurs.
     */
    public RPCOpaqueAuth getInitialAuthData()
    throws RPCError;

    /**
     * Get the "final" authentication information for the specified message
     * object.  This method allows the provider of an authentication generator
     * to create an authenticator that accounts for the full content of the
     * RPC message.  The returned data will then be used to overlay the area
     * of the message that holds the original data that was returned earlier
     * by the getInitialAuthData() method.  Note that the total size of the 
     * authentication information must be exactly the same as what was 
     * returned by the preceding getInitialAuthData() method invocation.
     * @param initial Initial authentication data used in the message.
     * @param msg Complete RPC message to be authenticated.
     * @return The final authentication information.
     * @exception RPCError If any RPC error occurs.
     */
    public RPCOpaqueAuth getFinalAuthData(RPCOpaqueAuth initial,
                                          XdrMessage msg)
    throws RPCError;
}
