package devmgr.api.rpc;

import devmgr.api.rpc.protocol.*;
import devmgr.api.runtime.*;

/**
 * Created with IntelliJ IDEA.
 * User: masonb
 * Date: 12/4/13
 * Time: 2:39 PM
 * To change this template use File | Settings | File Templates.
 */
public class XdrProtocolHelper {

    /**
     * This constant defines the offset in an RPC client request message
     * where the authentication information starts.  We need to define it
     * here so we can reposition to this offset and store the final version
     * of the message-dependent authentication data.
     */
    private static final int CLIENT_AUTH_OFFSET = 24;

    private ApiDefinition apiDef;
    private Decoder.TypeOverrideTable typeOverrides;

    /**
     * The following instance data tracks the RPC program and version
     * numbers of the server with which this client is interacting.
     */
    private int program;
    private int version;
    /**
     * Maintain a private null verifier object to ease the generation of
     * outbound request messages.  Also keep track of the authentication
     * generator object so we can conveniently invoke it during request
     * message creation.  The user can switch to an external authentication
     * generator if desired, but we provide a NULL generator by default.
     */
    private RPCAuthGenerator AuthGenerator;
    private RPCOpaqueAuth nullVerifier;
    private static TypeSpec RPCMsgTypeSpec;
    private static TypeSpec RPCOpaqueAuthTypeSpec;

    private static ApiDefinition RpcSpec;

    static
    {
        RpcSpec = new RPCProtocol();
        RPCMsgTypeSpec = RpcSpec.getTypeSpec(RPCMsg.class);
        RPCOpaqueAuthTypeSpec = RpcSpec.getTypeSpec(RPCOpaqueAuth.class);
    }


    public XdrProtocolHelper(ApiDefinition apiDef) {
        this.apiDef = apiDef;
        program = this.apiDef.getProgramNumber();
        version = this.apiDef.getProgramVersion();
        // Initialize a "null authenticator" object as a default.
        AuthGenerator = new RPCNullAuthGenerator();

        // Prepare a "null verifier" object for future use.  We currently
        // only support null verifiers; no true verifiers are used.
        nullVerifier = new RPCOpaqueAuth();
        nullVerifier.setType(RPCAuthType.AUTH_NONE);
        nullVerifier.setInfo("");


    }

    /**
     * Build a request message for the specified remote procedure call
     * interaction, and return the message to the caller for further
     * processing.
     * @param proc Procedure being called.
     * @param xid Transaction ID for this operation.
     * @param arg Argument being passed to the procedure.
     * @return A fully-constructed and ready-to-send message with CALL data.
     * @exception RPCError If any RPC error occurs.
     */
    public XdrMessage buildRequestMsg(int proc, int xid, Object arg)
            throws RPCError, EncodeException
    {
        ApiDefinition.ProcTableEntry pte = apiDef.getProcByNumber(proc);

        XdrMessage requestMsg = null;
        XdrEncoder encoder = new XdrEncoder();
        RPCMsg outMsg = new RPCMsg();
        RPCOpaqueAuth initialAuth = null;
        RPCOpaqueAuth finalAuth = null;

        // Populate the message elements.
        outMsg.setBody(new RPCMsgBody());
        outMsg.getBody().setCallMsg(new RPCCallMsg());

        // Set up the outbound message header.
        outMsg.setXid(xid);
        outMsg.getBody().setMsgType(RPCMsgType.CALL);
        outMsg.getBody().getCallMsg().setRpcVersion(RPCProtocol.VERSION);
        outMsg.getBody().getCallMsg().setProgramNo(program);
        outMsg.getBody().getCallMsg().setVersionNo(version);
        outMsg.getBody().getCallMsg().setProcedure(proc);

        // We only support NULL verifiers.
        outMsg.getBody().getCallMsg().setVerifier(nullVerifier);

        // Get the initial (placeholder) authentication data and include it.
        initialAuth = AuthGenerator.getInitialAuthData();
        outMsg.getBody().getCallMsg().setCredential(initialAuth);

        // Encode the header, followed by the argument data.
        encoder.encode(outMsg, RPCMsgTypeSpec);
        encoder.encode(arg, pte.argType);

        // Fetch the IOV-based representation request message data.
        requestMsg = encoder.getMessage();

        // Overlay the final authentication data onto the message.
        finalAuth = AuthGenerator.getFinalAuthData(initialAuth, requestMsg);
        encoder.setPosition(CLIENT_AUTH_OFFSET);
        encoder.encode(finalAuth, RPCOpaqueAuthTypeSpec);

        return requestMsg;
    }

    /**
     * Process a reply message for an RPC interaction with a server.  Check
     * for a valid transaction ID, check the reply status values, and then
     * decode the data into a result object to return to the caller.
     * @param proc Procedure number that was called.
     * @param replyMsg Reply message data.
     * @param xid Transaction ID for this operation.
     * @return Result object
     * @exception RPCError If any RPC error occurs.
     */
    public Object processReplyMsg(int proc, XdrMessage replyMsg, int xid)
            throws RPCError, DecodeException
    {
        ApiDefinition.ProcTableEntry pte = apiDef.getProcByNumber(proc);

        // Decode the reply message's RPC header.
        XdrDecoder decoder = new XdrDecoder(replyMsg);
        RPCMsg msg = (RPCMsg) decoder.decode(RPCMsgTypeSpec);

        // Check the header fields for any indication of error.
        if ( (msg.getXid() != xid) ||
                (msg.getBody().getMsgType() != RPCMsgType.REPLY) )
        {
            throw new RPCError(RPCError.ErrorType.PROTOCOL_ERROR);
        }

        if ( msg.getBody().getReplyMsg().getStatus() !=
                RPCReplyStatus.MSG_ACCEPTED )
        {
            throw new RPCError(RPCError.ErrorType.REQUEST_REJECTED);
        }

        if ( msg.getBody().getReplyMsg().getAcceptedReply().
                getVerifier().getType() !=
                RPCAuthType.AUTH_NONE )
        {
            throw new RPCError(RPCError.ErrorType.BAD_VERIFIER);
        }

        // Check here for indications of high-level failures on the
        // remote server system.
        switch ( msg.getBody().getReplyMsg().getAcceptedReply().
                getBody().getStatus() )
        {
            case SUCCESS:
                break;

            case PROG_UNAVAIL:
                throw new RPCError(RPCError.ErrorType.PROG_UNAVAILABLE);

            case PROG_MISMATCH:
                throw new RPCError(RPCError.ErrorType.PROG_MISMATCH);

            case PROC_UNAVAIL:
                throw new RPCError(RPCError.ErrorType.PROC_UNAVAILABLE);

            case GARBAGE_ARGS:
                throw new RPCError(RPCError.ErrorType.ARGS_REJECTED);

            case SYSTEM_ERR:
            default:
                throw new RPCError(RPCError.ErrorType.SYSTEM_ERROR);
        }

        // Decode the result value, which should be next in the
        // XDR stream (i.e. after the message header that we just
        // finished decoding).
        if (typeOverrides != null)
            decoder.setTypeOverrides(typeOverrides);

        return decoder.decode(pte.resultType);
    }
    public RPCAuthGenerator getAuthGenerator() {
        return AuthGenerator;
    }

    public void setAuthGenerator(RPCAuthGenerator AuthGenerator) {
        this.AuthGenerator = AuthGenerator;
    }

    public void setTypeOverrides(Decoder.TypeOverrideTable typeOverrides) {
        this.typeOverrides = typeOverrides;
    }
}
