package devmgr.api.runtime;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.LinkedList;

/**
 * This class implements a decoder for objects that are defined
 * with RPC-oriented data types (e.g. opaque byte arrays, discriminated
 * unions, fixed/variable-length arrays, etc.).  It is abstract, as it
 * requires a subclass that actually extracts encoded objects from a
 * particular representation type (e.g. JSON, XML, etc.).  The core logic
 * for driving the decoding operations resides here.
 */
public abstract class Decoder
{
    public static enum Context
    {
        NONE,
        ARRAY_FIRST,
        ARRAY_NONFIRST,
        STRUCT_FIRST,
        STRUCT_NONFIRST,
        UNION_FIRST,
        UNION_NONFIRST
    }
    
    protected String m_DefaultRootName = "root";
    
    @SuppressWarnings("serial")
    public static class TypeOverrideTable
    extends HashMap<Class<?>, Class<?>>
    {
    }

    public TypeOverrideTable m_TypeOverrideTable = new TypeOverrideTable();
        
    // These are all of the abstract methods that have to be
    // implemented by subclasses that are focused on decoding
    // from a particular representation.

    protected abstract TypeBinding
    getStructItem(String enclosingName, Context ctx, StructSpec spec)
    throws DecodeException;
    
    protected abstract String
    getUnionItem(String enclosingName, Context ctx, UnionSpec spec)
    throws DecodeException;
    
    protected abstract boolean
    isEndOfArray(String name, Context ctx)
    throws DecodeException;
    
    protected abstract byte 
    decodeByte(String name)
    throws DecodeException;
    
    protected abstract short 
    decodeShort(String name)
    throws DecodeException;
    
    protected abstract int 
    decodeInt(String name)
    throws DecodeException;
    
    protected abstract long 
    decodeLong(String name)
    throws DecodeException;
    
    protected abstract boolean 
    decodeBoolean(String name)
    throws DecodeException;
    
    protected abstract float 
    decodeFloat(String name)
    throws DecodeException;
    
    protected abstract double 
    decodeDouble(String name)
    throws DecodeException;
    
    protected abstract String 
    decodeString(String name, int maxLength)
    throws DecodeException;
    
    protected abstract String 
    decodeOpaque(String name, OpaqueSpec typeSpec)
    throws DecodeException;
    
    protected abstract void 
    decodeArrayStart(String name, Context ctx,
                     boolean isVariable, 
                     int maxLength)
    throws DecodeException;
    
    protected abstract void 
    decodeArrayEnd(String name,
                   boolean isVariable, 
                   int maxLength)
    throws DecodeException;
    
    protected abstract boolean 
    decodeStructStart(String name, Context ctx, StructSpec spec)
    throws DecodeException;
    
    protected abstract void 
    decodeStructEnd(String name, StructSpec spec)
    throws DecodeException;
    
    protected abstract boolean 
    decodeUnionStart(String name, Context ctx, UnionSpec spec)
    throws DecodeException;
    
    protected abstract void 
    decodeUnionEnd(String name, UnionSpec spec)
    throws DecodeException;
    
    protected abstract EnumValue 
    decodeEnum(String name, EnumSpec spec)
    throws DecodeException;

    /**
     * This method provides a default implementation for decoding 
     * enumeration sets.  If a subclass implements enumeration sets
     * as standard arrays of enumerations, they can just use this
     * base class implementation.  Otherwise, they need to override
     * it with a different implementation.
     * @param name Name of the enumeration set object.
     * @param ctx Decoding context.
     * @param spec TypeSpec for the enumeration itself.
     * @return An array of EnumValue objects.
     * @throws DecodeException
     */
    protected EnumValue[] 
    decodeEnumSet(String name, Context ctx, EnumSetSpec spec)
    throws DecodeException
    {
        LinkedList<EnumValue> values = new LinkedList<EnumValue>();
        
        decodeArrayStart(name, ctx, true, 0);
        
        Context innerCtx = Context.ARRAY_FIRST;
        while(! isEndOfArray(name, innerCtx))
        {
            EnumValue val = decodeEnum(name, spec.enumType);
            values.add(val);
            
            innerCtx = Context.ARRAY_NONFIRST;
        }
        
        decodeArrayEnd(name, true, 0);

        EnumValue[] result = new EnumValue[values.size()];
        return values.toArray(result);
    }

    /**
     * Sets the type override table for this decoder.
     * @param map Map containing base-class-to-sub-class entries.
     */
    public void
    setTypeOverrides(TypeOverrideTable table)
    {
        m_TypeOverrideTable = table;
    }
    
    /**
     * Decode an object of the specified type from the input stream.
     * @param type The type descriptor for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    public Object
    decode(TypeSpec type)
    throws DecodeException
    {
        return processByType(null, m_DefaultRootName, Context.NONE, type);
    }

    /**
     * Decode an object of the specified type from the input stream.
     * Note that the caller has to provide a reference to the table
     * that contains the type information for the desired class.
     * @param cls The class of the desired object.
     * @param api API in which the class is defined.
     * @return The decoded object.
     * @throws DecodeException
     */
    public Object
    decode(Class<?> cls, ApiDefinition api)
    throws DecodeException
    {
        return processByType(null, m_DefaultRootName, Context.NONE, 
                             api.getTypeSpec(cls));
    }
    
    /**
     * Process an object from the input stream, constructing an
     * appropriately-typed value.  Return the decoded object, for 
     * which all "embedded" data (e.g. structure fields, etc.) will
     * also be fully present and decoded.
     * @param name The name of the entity.
     * @param ctx The decoding context.
     * @param objType The entity's type information.
     */
    protected Object
    processByType(String name, Context ctx, TypeSpec objType)
    throws DecodeException
    {
        return processByType(null, name, ctx, objType);
    }
    
    /**
     * Process an object from the input stream, constructing an
     * appropriately-typed value.  Return the decoded object, for 
     * which all "embedded" data (e.g. structure fields, etc.) will
     * also be fully present and decoded.
     * @param obj The object to be "filled", or null to instantiate a new one.
     * @param name The name of the entity.
     * @param ctx The decoding context.
     * @param objType The entity's type information.
     */
    protected Object
    processByType(Object obj, String name, Context ctx, TypeSpec objType)
    throws DecodeException
    {
        // Process based on object type.
        switch (objType.typeCode)
        {
            case BYTE:
                return decodeByte(name);
                
            case CHAR:
                return (char) decodeByte(name);
                
            case SHORT:
                return decodeShort(name);
             
            case INT:
                return decodeInt(name);
                
            case LONG:
                return decodeLong(name);
                
            case BOOLEAN:
                return decodeBoolean(name);
                
            case FLOAT:
                return decodeFloat(name);
                
            case DOUBLE:
                return decodeDouble(name);
                
            case STRING:
                return processString(name, (StringSpec) objType);
                
            case OPAQUE:
                return processOpaque(name, (OpaqueSpec) objType);
                
            case ARRAY:
                return processArray(name, ctx, (ArraySpec) objType);
                
            case ENUM:
                return processEnum(name, (EnumSpec) objType);
                
            case ENUMSET:
                return processEnumSet(name, ctx, (EnumSetSpec) objType);
                
            case POINTER:
                return processByType(obj, name, ctx,
                                     ((PointerSpec) objType).targetType);
                
            case STRUCT:
                return processStruct(obj, name, ctx, (StructSpec) objType);
                
            case TYPEDEF:
                return processTypedef(obj, name, ctx, (TypedefSpec) objType);
                
            case SUBCLASS:
                return processSubclass(obj, name, ctx, (SubclassSpec) objType);
                
            case UNION:
                return processUnion(obj, name, ctx, (UnionSpec) objType);
        }
        
        return null;
    }

    /**
     * This method processes a string.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    private String 
    processString(String name, StringSpec typeSpec)
    throws DecodeException
    {
        return decodeString(name, typeSpec.maxLength);
    }
    
    /**
     * This method processes an opaque byte array.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    private String 
    processOpaque(String name, OpaqueSpec typeSpec)
    throws DecodeException
    {
        return decodeOpaque(name, typeSpec);
    }
    
    /**
     * This method processes an enumeration.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    @SuppressWarnings({ "unchecked", "rawtypes" })
    private Object
    processEnum(String name, EnumSpec typeSpec)
    throws DecodeException
    {
        // Decode from the input stream.
        EnumValue enumVal = decodeEnum(name, typeSpec);
        
        // Construct an appropriately-typed result object.
        return Enum.valueOf((Class<Enum>) typeSpec.cls, enumVal.name);
    }

    /**
     * This method processes an enumeration set.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    @SuppressWarnings({ "unchecked", "rawtypes" })
    private Object
    processEnumSet(String name, Context ctx, EnumSetSpec typeSpec)
    throws DecodeException
    {
        // Decode from the input stream.
        EnumValue[] values = decodeEnumSet(name, ctx, typeSpec);
        
        Object result = 
            Array.newInstance(typeSpec.enumType.cls, values.length);
        
        // Construct array elements as appropriately-typed objects.
        for (int i = 0; i < values.length; i++)
        {
            Array.set(result,  i, 
                      Enum.valueOf((Class<Enum>) typeSpec.enumType.cls, 
                                   values[i].name));
        }
        
        return result;
    }

    /**
     * This method processes a subclass.  Note that subclasses are
     * always simple aliases of a base class, with no additional
     * data fields.  We handle them by instantiating (if necessary)
     * a new subclass object, but then feeding that object to the 
     * decode process for the base class.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    private Object
    processSubclass(Object obj, String name, Context ctx, 
                    SubclassSpec typeSpec)
    throws DecodeException
    {
        try
        {
            Object result = null;
            if (obj != null)
                result = obj;
            else if (typeSpec.cls != null)
                result = determineClass(typeSpec.cls).newInstance();

            return processByType(result, name, ctx, typeSpec.superType);
        }
        catch(Exception caught)
        {
            throw buildException(caught, typeSpec.cls);
        }
    }

    /**
     * This method processes a typedef.  Typedefs are only used in cases
     * where a base type definition is being extended to an array or 
     * pointer context.  The resulting type will have a single data field
     * named "value", with either pointer or array extensions for it.  We
     * handle such cases by instantiating a new object, then invoking the
     * decode process for the embedded value field.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    private Object
    processTypedef(Object obj, String name, Context ctx, TypedefSpec typeSpec)
    throws DecodeException
    {
        try
        {
            Object value = processByType(null, name, ctx, typeSpec.typeSpec);
            
            Object result = 
                (obj != null) ? obj : 
                    determineClass(typeSpec.cls).newInstance();
            
            Field fld = typeSpec.cls.getDeclaredField("value");
            fld.setAccessible(true);
            fld.set(result, value);
            
            return result;
        }
        catch(Exception caught)
        {
            throw buildException(caught, typeSpec.cls);
        }
    }

    /**
     * This method processes an array.  It emits up-calls to the
     * methods that bracket the start and end of an array.  In
     * between those two, it emits per-element calls to decode
     * the array items themselves.
     * @param m_Name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    private Object
    processArray(String arrayName, Context ctx, ArraySpec typeSpec)
    throws DecodeException
    {
        LinkedList<Object> items = new LinkedList<Object>();
        
        decodeArrayStart(arrayName, ctx, 
                         typeSpec.isVariable, typeSpec.maxLength);

        // Accumulate decoded items in a linked list.
        int actualLength = 0;
        Context itemContext = Context.ARRAY_FIRST;
        while (true)
        {
            if (isEndOfArray(arrayName, itemContext))
                break;
            
            Object obj = processByType(null, arrayName, itemContext, 
                                       typeSpec.elementType);
            itemContext = Context.ARRAY_NONFIRST;
            
            if (obj == null)
                break;
            
            if ( (typeSpec.maxLength > 0) && 
                 (actualLength == typeSpec.maxLength) )
            {
                throw DecodeException.build
                        ("%s[]: max length exceeded",
                         typeSpec.elementClass.getCanonicalName());
            }
                
            items.addLast(obj);
            actualLength++;
        }            

        // Verify length of fixed-length arrays.
        if ( (! typeSpec.isVariable) && (actualLength != typeSpec.maxLength) )
        {
            throw DecodeException.build
                    ("%s: length %d, should be %d",
                     typeSpec.elementClass.getCanonicalName(), 
                     actualLength, typeSpec.maxLength);
        }
        
        decodeArrayEnd(arrayName, typeSpec.isVariable, typeSpec.maxLength);
        
        // Construct the result array and set its elements to the
        // objects we accumulated in our linked list.
        Object result = 
            Array.newInstance(determineClass(typeSpec.elementClass), 
                              actualLength);
        
        for (int i = 0; i < actualLength; i++)
        {
            Array.set(result, i, items.get(i));
        }
        
        return result;
    }

    /**
     * This method decodes a structure.  It emits up-calls to the
     * methods that bracket the start and end of the structure.
     * In between those two, it emits per-field calls to decode
     * the structure fields themselves.
     * @param name Name of the entity.
     * @param ctx Decoding context.
     * @param typeSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    private Object
    processStruct(Object obj, String name, Context ctx, StructSpec objType)
    throws DecodeException
    {
        try
        {
            Object result;
            
            Boolean hasContent = decodeStructStart(name, ctx, objType);
            
            // If there is no content available for this structure,
            // don't try to process it at all.
            if (! hasContent)
            {
                result = null;
            }
            
            // If this is a "collapsed" structure, we just want to make
            // it look like the structure wrappings are not there.  The
            // "struct" is really more like an alias for its one/only
            // defined field.
            else if (objType.isCollapsed)
            {
                result = processByType(name, ctx, objType.fields[0].typeSpec);
            }
            
            // For non-collapsed structures, process field-by-field.
            else
            {
                result = (obj != null) ? obj : 
                    determineClass(objType.cls).newInstance();
            
                Context itemContext = Context.STRUCT_FIRST;
                while (true)
                {
                    // Fetch the next field in the input stream.
                    TypeBinding binding = 
                        getStructItem(name, itemContext, objType);
    
                    itemContext = Context.STRUCT_NONFIRST;
                    
                    // Check for end of structure.
                    if (binding == null)
                        break;
                    
                    Object fldValue = 
                        processByType(binding.itemName, 
                                      itemContext, binding.typeSpec);
                                
                    Field fld = objType.cls.
                                    getDeclaredField(binding.itemName);
                    fld.setAccessible(true);
                    fld.set(result, fldValue);
                }
            }
        
            decodeStructEnd(name, objType);
            
            return result;
        }
        catch (Exception caught)
        {
            throw DecodeException.build(caught, "Processing struct '%s'",
                                        objType.cls.getCanonicalName());
        }
    }
            
    /**
     * This method decodes a union.  It emits up-calls to the
     * methods that bracket the start and end of the union, then
     * handles decoding of the discriminator and the data value.
     * Note that we have to allow the discriminator and the data
     * value to appear in any order, so that makes it a bit more
     * challenging.  In particular, we may encounter the data value
     * before we know the discriminator, so we use the name of the
     * data field as our key for determining the data type.  In
     * that case, we have to check the discriminator that comes
     * later to ensure that it matches the data type we actually
     * encountered.
     * @param m_Name Name of the entity.
     * @param ctx Decoding context.
     * @param unionSpec Type specification for the object.
     * @return The decoded object.
     * @throws DecodeException
     */
    @SuppressWarnings({ "unchecked", "rawtypes" })
    private Object
    processUnion(Object obj, String unionName, Context ctx, UnionSpec unionSpec)
    throws DecodeException
    {
        String clsName = unionSpec.cls.getCanonicalName();
        
        try
        {
            boolean hasContent = decodeUnionStart(unionName, ctx, unionSpec);
            
            // Do a quick exit, with a null result, if there is no content
            // for this union.
            if (! hasContent)
            {
                decodeUnionEnd(unionName, unionSpec);
                return null;
            }
            
            EnumValue discrim = null;
            TypeBinding expectedAltBinding = null;
            TypeBinding actualAltBinding = null;
            Object altValue = null;

            EnumSpec enumSpec = (EnumSpec) unionSpec.discrimType;
            
            // Determine the class to use for instantiation.  Note that
            // the discriminator may be either an enumeration, or a
            // subclass (of a subclass of a ...) of an enumeration. 
            Class<?> discrimClass;
            if (unionSpec.discrimType instanceof SubclassSpec)
                discrimClass = ((SubclassSpec) unionSpec.discrimType).cls;
            else
                discrimClass = enumSpec.cls;
            
            // Iteratively extract "fields" from the input stream.
            // Note that we do NOT impose a constraint that the
            // discriminator appears BEFORE the alternate, so we
            // have to be flexible in how we handle orderings.
            Context itemContext = Context.UNION_FIRST;
            while (true)
            {
                // Fetch the next field name within this union.
                String fieldName = 
                    getUnionItem(unionName, itemContext, unionSpec);
                
                itemContext = Context.UNION_NONFIRST;
                
                // Check for end of union.
                if (fieldName == null)
                    break;
            
                // If we haven't seen the discriminator yet, check
                // for it here.
                if ( (discrim == null) &&
                     (fieldName.equals(unionSpec.discrimName)) )
                {
                    // Decode from the input stream.
                    discrim = decodeEnum(unionName, enumSpec);

                    // Determine the expected alternate, given this 
                    // discriminator value.  If there's no mapping that
                    // matches, use the default (which may be null, 
                    // i.e. 'void').
                    expectedAltBinding = unionSpec.caseMap.get(discrim.value);
                    if (expectedAltBinding == null)
                        expectedAltBinding = unionSpec.defaultBinding;
                }
                
                // We didn't get the discriminator, so we should have an
                // alternate instead.  Make sure there's only one!
                else if (actualAltBinding != null)
                {
                    throw DecodeException.build("%s: multiple alternates",
                                                clsName);
                }
                
                // Now we can decode the alternate.
                else 
                {
                    // Check for default first, then for a normal alternate.
                    if ( (unionSpec.defaultBinding != null) &&
                         (fieldName.equals(unionSpec.defaultBinding.itemName)) )
                        actualAltBinding = unionSpec.defaultBinding;
                    else
                        actualAltBinding = unionSpec.nameMap.get(fieldName);
                    
                    // If no binding was found, we can't proceed.
                    if (actualAltBinding == null)
                        throw DecodeException.build
                                ("%s: bad alternate name '%s'",
                                 clsName, fieldName);

                    // Extract the value, based on its type.
                    if (actualAltBinding.typeSpec != null)
                        altValue = processByType(fieldName, itemContext, 
                                                 actualAltBinding.typeSpec);
                }
            }
            
            // Make sure we got a discriminator, and that the alternate
            // we would EXPECT with that discriminator matches the one
            // that was actually PRESENT in the input stream.
            if (discrim == null)
                throw DecodeException.build("%s: No discriminator",
                                            clsName);
            if (actualAltBinding != expectedAltBinding)
                throw DecodeException.build
                        ("%s: Discriminator/alternate mismatch",
                         clsName);

            // Now we just need to construct the result object, and fill in
            // the discriminator and alternate values.
            Object result = (obj != null) ? obj : 
                determineClass(unionSpec.cls).newInstance();
            
            // Add the discriminator to the main result.
            Field fld = unionSpec.cls.getDeclaredField(unionSpec.discrimName);
            fld.setAccessible(true);
            fld.set(result, 
                    Enum.valueOf((Class<Enum>) discrimClass, discrim.name));
            
            // Add the alternate value to the main result (unless we're
            // dealing with a 'void' case).
            if (actualAltBinding != null)
            {
                try{
                    fld = unionSpec.cls.getDeclaredField(actualAltBinding.itemName);
                    fld.setAccessible(true);
                    fld.set(result, altValue);
                }catch (NoSuchFieldException e){
                    //bdm, left this asset in so we can find this on unit tests but I need this to not blow up for now.
                    assert (false);
                }
            }
            
            decodeUnionEnd(unionName, unionSpec);
            
            return result;
        }
        catch (Exception caught)
        {
            throw DecodeException.build(caught, "Processing union '%s'",
                                        clsName);
        }
    }
     
    /**
     * This method find the class to use for instantiation, given the
     * specified starting point.  This is where we determine if we should
     * use a subclass, instead of the starting point, to enable use of
     * "augmented" derived classes.
     * @param cls The starting-point class.
     * @return The class to use.
     */
    private Class<?>
    determineClass(Class<?> cls)
    {
        Class<?> override = m_TypeOverrideTable.get(cls);
        
        if (override != null)
            return override;
        else
            return cls;
    }
    
    /**
     * This method builds an exception to be thrown back to the original
     * caller, including information about the class being processed.
     * @param ex Original exception that was caught.
     * @param cls Object type being processed.
     * @return An DecodeException.
     */
    private DecodeException
    buildException(Exception ex, Class<?> cls)
    {
        String clsName = (cls == null) ? "<None>" : cls.getCanonicalName();
        return DecodeException.build(ex, "Processing class '%s'", clsName);
    }
}