package devmgr.api.runtime;

import java.io.Reader;

/**
 * This class implements a decoder for JSON input streams.  It provides
 * methods for processing such input data to construct Java objects
 * that coincide with the data in a stream.
 */
public class JsonDecoder
extends Decoder
{
    private Reader m_In;
    private char m_PeekChar;
    private boolean m_Peeked = false;
    
    private static final int DEFAULT_MAX_STRING_LEN = 1024;
    private static final String TOKEN_TERMINATORS = "{}[]:,\"\t\n\r ";

    /**
     * Constructor.
     * @param rdr Input reader for stream data to be decoded.
     */
    public
    JsonDecoder(Reader rdr)
    {
        m_In = rdr;
    }

    /**
     * This method peeks at (but does NOT consume) the next available 
     * character in the input stream.
     * @return The next available character.
     * @throws DecodeException
     */
    private char
    peekChar()
    throws DecodeException
    {
        try
        {
            if (! m_Peeked)
            {
                m_PeekChar = (char) m_In.read();
                m_Peeked = true;
            }
        }
        catch (Exception ex)
        {
            throw DecodeException.build(ex, "JSON: peek failed");
        }
        
        return m_PeekChar;
    }
    
    /**
     * This method returns (and consumes) the next available character
     * in the input stream.  If a character was "peeked at" just prior
     * to this call, that character is returned (and consumed).  If not,
     * we just go ahead and fetch the next available character.
     * @return The next available character.
     * @throws DecodeException
     */
    private char
    fetchChar()
    throws DecodeException
    {
        try
        {
            if (m_Peeked)
            {
                m_Peeked = false;
                return m_PeekChar;
            }
            
            return (char) m_In.read();
        }
        catch (Exception ex)
        {
            throw DecodeException.build(ex, "JSON: fetch failed");
        }
    }
    
    /**
     * This method skips the specified number of characters in the 
     * input stream.  It accounts for a potential peek operation that
     * preceded this call.
     * @param n Number of characters to skip.
     * @throws DecodeException
     */
    private void
    skip(int n)
    throws DecodeException
    {
        try
        {
            while (n > 0)
            {
                if (m_Peeked)
                    m_Peeked = false;
                else
                    m_In.read();

                n--;
            }
        }
        catch (Exception ex)
        {
            throw DecodeException.build(ex, "JSON: skip failed");
        }
    }
    
    /**
     * Peek at a non-white-space character from the input stream.  We skip
     * over (i.e. ignore) all white space leading up to it, and then PEEK
     * (without consuming) the non-white character.
     * @return The next non-white character in the input stream.
     */
    private char
    peekNonWhite()
    throws DecodeException
    {
        char x;

        while (true)
        {
            x = peekChar();

            if ( (x == ' ') || (x == '\t') || (x == '\n') || (x == '\r') )
            {
                skip(1);
                continue;
            }

            return x;
        }
    }

    /**
     * Get a non-white-space character from the input stream.  We skip over
     * (i.e. ignore) all white space leading up to it.
     */
    private char 
    fetchNonWhite()
    throws DecodeException
    {
        char x = peekNonWhite();
        skip(1);
        return x;
    }

    /**
     * Fetch an integer value from the input stream, and return it as
     * a 64-bit object to the caller.
     * @return Fetched integer, as a long value.
     */
    private long
    fetchInt()
    throws DecodeException
    {
        String str = fetchToken();
        long result = 0;
        
        try
        {
            result = Long.parseLong(str);
        }
        catch (Exception ex)
        {
            throw DecodeException.build
                    ("JSON: bad integer: %s", str);
        }
        
        return result;
    }

    /**
     * Fetch a non-quoted token, interpreting any white space or delimiter
     * character as a string terminator.  The caller's buffer is filled with
     * the extracted token, including a zero-terminator.
     */
    private String 
    fetchToken()
    throws DecodeException
    {
        StringBuilder buf = new StringBuilder();
        
        // Start by skipping over any white space.
        peekNonWhite();
        
        // Now accumulate non-white-space characters.
        while (true)
        {
            
            // Peek at the next character, checking for end-of-input.
            char x = peekChar();
            if (x == 0)
                break;
            
            // If we hit a terminator, stop accumulating.
            else if (stringContains(TOKEN_TERMINATORS, x) != -1)
                break;

            // Else consume the character and keep going.
            else
            {
                buf.append(x);
                skip(1);
            }
        }
        
        // If we got a zero-length token, it's a problem.
        if (buf.length() == 0) 
            throw new DecodeException("JSON: bad token");

        return buf.toString();
    }

    /**
     * Get a (double) floating point value from the input stream.
     * @return (Double) floating point value.
     * @throws DecodeException
     */
    private double
    fetchDouble()
    throws DecodeException
    {
        String s = fetchToken();
        double result = 0;
        
        try
        {
            result = Double.parseDouble(s);
        }
        catch (Exception ex)
        {
            throw DecodeException.build
                    ("JSON: bad floating point: %s", s);
        }
        
        return result;
    }
    
    /**
     * Consume the specified character in the input stream (and also any
     * white space leading up to it).  If that character is NOT the first
     * non-white-space character encountered, it's considered an error.
     * @param desiredChar Desired character to consume.
     */
    private void 
    consumeChar(char desiredChar)
    throws DecodeException
    {
        char x;
        
        x = fetchNonWhite();
        if (x != desiredChar)
            throw new DecodeException("JSON: unexpected character");
    }

    /**
     * See if the specified string contains the give character.
     * @param str String to be searched.
     * @param c Character of interest.
     * @return Index of character, or -1 if not found.
     */
    private static int
    stringContains(String str, char c)
    {
        for (int i = 0; i < str.length(); i++)
            if (str.charAt(i) == c)
                return i;
        
        return -1;
    }
    
    /**
     * Decode a string.
     * @param maxLen Maximum number of characters allowed, or 0 for unlimited.
     * @return The decoded string, without quotes.
     * @throws DecodeException
     */
    private String
    getString(int maxLen)
    throws DecodeException
    {
        String exMsg = "JSON: bad string";
        StringBuilder sb = new StringBuilder();
        
        if (fetchChar() != '"')
            throw new DecodeException(exMsg);
        
        int len = 0;
        char c;
        while ( (c = fetchChar()) != '"' )
        {
            if ( (maxLen > 0) && (len >= maxLen) )
                throw new DecodeException(exMsg);
            
            sb.append(c);
            len++;
        }
        
        return sb.toString();
    }
    
    /**
     * Decode a string (no max length specified, so use default).
     * @return The decoded string, without quotes.
     * @throws DecodeException
     */
    private String
    getString()
    throws DecodeException
    {
        return getString(DEFAULT_MAX_STRING_LEN);
    }
    
    /**
     * This method gets the next name (discriminator OR data field) in
     * the input stream for the specified union.  If we reach the end
     * of the union, a null value is returned.
     */
    @Override
    protected String 
    getUnionItem(String enclosingName, Context ctx, UnionSpec objType)
    throws DecodeException
    {
        char c = peekNonWhite();

        if (c == '}')
            return null;

        if ( (ctx == Context.UNION_NONFIRST) && (c == ',') )
            skip(1);
        
        String name = getString();
        consumeChar(':');
        return name;
    }

    /**
     * This method gets the type binding information for the next
     * field in the input stream for the specified structure.  If we
     * reach the end of the structure, a null value is returned.
     */
    @Override
    protected TypeBinding
    getStructItem(String enclosingName, Context ctx, StructSpec objType)
    throws DecodeException
    {
        char c = peekNonWhite();

        if (c == '}')
            return null;

        if ( (ctx == Context.STRUCT_NONFIRST) && (c == ',') )
            skip(1);
        
        String name = getString();
        consumeChar(':');
        
        TypeBinding binding = objType.nameMap.get(name);
        
        if (binding == null)
        {
            throw DecodeException.build
                    ("%s: bad field name: %s",
                     objType.cls.getCanonicalName(), name);
        }
        
        return binding;
    }

    @Override
    protected boolean
    isEndOfArray(String name, Context ctx)
    throws DecodeException
    {
        char c = peekNonWhite();
        
        if (c == ']')
            return true;
    
        if (ctx == Context.ARRAY_NONFIRST)
            consumeChar(',');
    
        return false;
    }

    @Override
    protected byte decodeByte(String name)
    throws DecodeException
    {
        long result = fetchInt();
        return (byte) result;
    }

    @Override
    protected short decodeShort(String name)
    throws DecodeException
    {
        long result = fetchInt();
        return (short) result;
    }

    @Override
    protected int decodeInt(String name)
    throws DecodeException
    {
        long result = fetchInt();
        return (int) result;
    }

    @Override
    protected long decodeLong(String name)
    throws DecodeException
    {
        consumeChar('"');
        long result = fetchInt();
        consumeChar('"');
        
        return result;
    }

    @Override
    protected boolean decodeBoolean(String name)
    throws DecodeException
    {
        String token = fetchToken();
        
        if (token.equals("true"))
            return true;
        else if (token.equals("false"))
            return false;
        
        throw new DecodeException("JSON: bad boolean");
    }

    @Override
    protected float decodeFloat(String name)
    throws DecodeException
    {
        return (float) fetchDouble();
    }

    @Override
    protected double decodeDouble(String name)
    throws DecodeException
    {
        return fetchDouble();
    }

    @Override
    protected String decodeString(String name, int maxLength)
    throws DecodeException
    {
        return getString(maxLength);
    }

    @Override
    protected String 
    decodeOpaque(String name, OpaqueSpec typeSpec)
    throws DecodeException
    {
        String result = getString(typeSpec.maxLength * 2);
        
        if ((result.length() % 2) != 0)
            throw new DecodeException("JSON: bad opaque data");
        
        if ( (! typeSpec.isVariable) && 
             ((result.length() / 2) != typeSpec.maxLength) )
            throw new DecodeException("JSON: bad fixed opaque length");
        
        return result;
    }

    @Override
    protected void 
    decodeArrayStart(String name, Context ctx,
                     boolean isVariable, int maxLength)
    throws DecodeException
    {
        consumeChar('[');
    }

    @Override
    protected void 
    decodeArrayEnd(String name, boolean isVariable, int maxLength)
    throws DecodeException
    {
        consumeChar(']');
    }

    @Override
    protected boolean 
    decodeStructStart(String name, Context ctx, StructSpec spec)
    throws DecodeException
    {
        // Do nothing for collapsed structures.
        if (spec.isCollapsed)
            return true;
        
        consumeChar('{');
        return true;
    }

    @Override
    protected void 
    decodeStructEnd(String name, StructSpec spec)
    throws DecodeException
    {
        // Do nothing for collapsed structures.
        if (spec.isCollapsed)
            return;
        
        consumeChar('}');
    }

    @Override
    protected boolean 
    decodeUnionStart(String name, Context ctx, UnionSpec spec)
    throws DecodeException
    {
        consumeChar('{');
        return true;
    }

    @Override
    protected void 
    decodeUnionEnd(String name, UnionSpec spec)
    throws DecodeException
    {
        consumeChar('}');
    }

    @Override
    protected EnumValue 
    decodeEnum(String name, EnumSpec typeSpec)
    throws DecodeException
    {
        String label = getString();
        
        Integer value = typeSpec.nameToValueMap.get(label);
        if (value == null)
            throw DecodeException.build
                    ("JSON: bad enum value %s for %s in %s",
                     value, typeSpec.cls.getCanonicalName(), name);
        
        return new EnumValue(label, value);
    }
}