package devmgr.api.runtime;

import java.io.IOException;
import java.io.OutputStream;

/**
 * This class provides XdrEncoder capability for an output stream
 * object.  I.e. the encoded data is written to an OutputStream.
 */
public class XdrStreamEncoder
extends XdrEncoder
{
    private boolean m_Flushed = false;
    private boolean m_Closed = false;
    private OutputStream m_Out;
    
    /**
     * Constructor.  Sets up an encoder so that its eventual
     * encoded data can be written to the specified stream.
     * @param os Output stream to which we'll write the data.
     * @throws DecodeException
     */
    public XdrStreamEncoder(OutputStream os)
    {
        m_Out = os;
    }

    /**
     * Flush all accumulated encoded data to the stream.  This method
     * MUST be called in order to push the encoded data out from its
     * buffered form to the stream object.  It should ONLY be called
     * one time, at the end of the encoding process.
     * @throws EncodeException
     */
    @Override
    protected void
    flush()
    throws EncodeException
    {
        if (m_Flushed)
            throw EncodeException.build("Multiple flush() calls not allowed");
        
        m_Flushed = true;
        XdrMessage msg = getMessage();
        
        try
        {
            int residual = msg.getMessageSize();
            int bufNum = 0;
            while (residual > 0)
            {
                byte[] buf = msg.getBuffer(bufNum++);
                int amt = Math.min(residual, buf.length);
                
                if (amt == buf.length)
                    m_Out.write(buf);
                else
                    m_Out.write(buf, 0, amt);
                
                residual -= amt;
            }
            
            m_Out.flush();
        }
        
        catch (Exception ex)
        {
            throw EncodeException.build(ex, "Can't write output stream");
        }
    }

    /**
     * Close the associated output stream.  The caller can either use this
     * method, or can close the stream directly via a call to the stream's
     * close() method.
     * @throws IOException, EncodeException
     */
    public void
    close()
    throws IOException, EncodeException
    {
        if (m_Closed)
            return;
        
        if (! m_Flushed)
            flush();
        
        m_Closed = true;
        m_Out.close();
    }
}